import sys
from ctypes import cdll

if sys.platform != 'darwin':
    raise OSError('libcompression can only be used on macOS')

try:
    _lib = cdll.LoadLibrary('libcompression.dylib')
except OSError as error:
    raise OSError('libcompression.dylib not found') from error



from importlib.metadata import version

from loguru import logger as _logger

from .compress import compress, decompress  # noqa: F401
from .types import Algorithm  # noqa: F401

__version__ = version(__package__)
_logger.disable(__package__)



import sys
from typing import BinaryIO

import click
from loguru import logger

from apple_compress import Algorithm, __version__, compress, decompress

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.command(context_settings=CONTEXT_SETTINGS)
@click.version_option(message=f'acompress {__version__}')
@click.option(
    '-i',
    '--input',
    'input_',
    type=click.File('rb'),
    help='Input file.',
    required=True,
)
@click.option(
    '-o',
    '--output',
    'output',
    type=click.File('wb'),
    help='Output file.',
    required=True,
)
@click.option(
    '-c', '--compress', 'action_type', flag_value='compress', help='Compress the data.'
)
@click.option(
    '-d',
    '--decompress',
    'action_type',
    flag_value='decompress',
    help='Decompress the data.',
)
@click.option(
    '-l',
    '--lzfse',
    'compression_type',
    flag_value='LZFSE',
    help='LZFSE compress the data.',
)
@click.option(
    '-b',
    '--lzfse-iboot',
    'compression_type',
    flag_value='LZFSE_IBOOT',
    help='LZFSE compress the data (iBoot compatible).',
)
@click.option(
    '-z',
    '--zlib',
    'compression_type',
    flag_value='ZLIB',
    help='zlib compress the data.',
)
@click.option(
    '-v',
    '--verbose',
    'verbose',
    is_flag=True,
    help='Increase verbosity.',
)
def main(
    input_: BinaryIO,
    output: BinaryIO,
    action_type: str,
    compression_type: str,
    verbose: bool,
) -> None:
    """A Python CLI tool for compression using Apple's libcompression."""

    if verbose:
        logger.remove()
        logger.add(
            sys.stderr,
            level='DEBUG',
            format='[{time:MMM D YYYY - hh:mm:ss A zz}] {level} {module}:{line} {message}',
        )
        logger.enable(__package__)
    else:
        sys.tracebacklimit = 0

    if sys.platform != 'darwin':
        click.echo('[ERROR] Only Darwin systems are supported. Exiting.')
        return

    if compression_type is None:
        raise click.BadParameter('Must specify a compression type.')

    if verbose:
        click.echo(f'Using compression type: {compression_type}')
    compression_type = getattr(Algorithm, compression_type)

    if action_type is None:
        raise click.BadParameter('Must specify either --compress or --decompress.')
    elif action_type == 'compress':
        output.write(compress(input_.read(), compression_type))
    elif action_type == 'decompress':
        output.write(decompress(input_.read(), compression_type))

    click.echo(
        f'{action_type.capitalize()}ed {compression_type.name} data to {output.name}.'
    )


if __name__ == '__main__':
    main()



from enum import Enum


class Algorithm(Enum):
    """
    Enum representing different compression algorithms.

    Attributes:
        LZ4 (int): LZ4 compression algorithm.
        LZ4_RAW (int): LZ4 compression algorithm (no block headers).
        ZLIB (int): ZLIB compression algorithm.
        LZMA (int): LZMA compression algorithm.
        LZFSE (int): LZFSE compression algorithm.
        LZFSE_IBOOT (int): LZFSE compression algorithm (iBoot-compatible).
    """

    LZ4 = 0x100
    LZ4_RAW = 0x101
    ZLIB = 0x205
    LZMA = 0x306
    LZFSE = 0x801
    LZFSE_IBOOT = 0x891
    # These are only supported on newer libcompression versions, disabled for now.
    # LZBITMAP = 0x702
    # BROTLI = 0xB02



from ctypes import create_string_buffer
from typing import Optional

from loguru import logger

from ._lib import _lib
from .errors import CompressionError
from .types import Algorithm


def compress(data: bytes, algorithm: Algorithm) -> bytes:
    """
    Compresses the given data using the specified algorithm.

    Args:
        data (bytes): The data to be compressed.
        algorithm (Algorithm): The compression algorithm to use.

    Returns:
        bytes: The compressed data.

    Raises:
        CompressionError: If the compression fails.
    """
    logger.info(
        f'Compressing data ({len(data)} bytes) with algorithm: {algorithm.name}'
    )
    logger.debug(f'Creating buffer of size: {len(data) + 256}')
    dest_buf = create_string_buffer(len(data) + 256)

    logger.debug('Calling compression_encode_buffer')
    size = _lib.compression_encode_buffer(
        dest_buf, len(data), data, len(data), None, algorithm.value
    )
    if size == 0:
        raise CompressionError('Failed to compress data')

    logger.success(
        f'Compression with algorithm: {algorithm.name} successful, returning {size} bytes'
    )
    return dest_buf.raw[:size]


def decompress(
    data: bytes, algorithm: Algorithm, decmp_size: Optional[int] = None
) -> bytes:
    """
    Decompresses the given data using the specified algorithm.

    Args:
        data (bytes): The compressed data to be decompressed.
        algorithm (Algorithm): The compression algorithm to use for decompression.
        decmp_size (Optional[int]): The expected size of the decompressed data. If not provided, it is twice the size of `data`. Providing this value is recommended on resource-limited systems.

    Returns:
        bytes: The decompressed data.

    Raises:
        CompressionError: If the decompression fails.
    """
    logger.info(
        f'Decompressing data ({len(data)} bytes) with algorithm: {algorithm.name}'
    )

    if decmp_size is None:
        logger.debug(
            f'No decompressed size provided, assuming data length * 2 ({len(data) * 2})'
        )
        decmp_size = len(data) * 2

    while True:
        logger.debug(f'Creating buffer of size: {decmp_size}')
        dest_buf = create_string_buffer(decmp_size)

        logger.debug('Calling compression_decode_buffer')
        size = _lib.compression_decode_buffer(
            dest_buf, decmp_size, data, len(data), None, algorithm.value
        )
        if size == 0:
            raise CompressionError('Failed to decompress data')

        if size == decmp_size:
            logger.debug(
                'Buffer too small for decompressed data, running again with 1.5x larger buffer'
            )
            decmp_size = round(decmp_size * 1.5)
            continue

        logger.success(
            f'Decompression with algorithm: {algorithm.name} successful, returning {size} bytes'
        )
        return dest_buf.raw[:size]
